class Dom < Xenodot::Migration
  def self.up
    execute %q{
      create or replace function xenodot.parent_hierarchy(child_hierarchy text) returns text as $ruby$
        return nil if child_hierarchy == ''
        child_hierarchy.scan(/\~*\w/)[0...-1].to_s
      $ruby$ language plruby immutable;
    }
    
    create_table "xenodot.node_head" do |t|
      t.column  "root_id",        :int4, :null => false
      t.column  "hierarchy",      :text, :null => false
      t.column  "name_id",        :int4, :null => false
    end
    
    execute %q{
      create or replace view xenodot.node as select * from xenodot.node_head;
    }
    
    execute <<-SQL
      alter table xenodot.node_head alter column hierarchy set storage main;
    
      alter table xenodot.node_head add foreign key(root_id) references xenodot.uri   initially deferred;
      alter table xenodot.node_head add foreign key(name_id) references xenodot.qname initially deferred;
      
      create unique index dom_structure on xenodot.node_head(root_id, hierarchy);
      cluster dom_structure on xenodot.node_head;

      create index parent on xenodot.node_head(xenodot.parent_hierarchy(hierarchy));
      create index name_speed on xenodot.node_head(name_id);
    SQL
    
    execute <<-SQL
      create or replace function xenodot.node(int4) returns xenodot.node as $sql$
        select * from xenodot.node_head where id = $1;
      $sql$ language sql stable;
    SQL
    
    execute %q{
      create or replace function xenodot.ancestor_hierarchy(val text) returns text[] as $ruby$
        return [''] if val == ''
        elements = val.scan(/~*\w/)
        result = ['']
        elements.size.times{|ii| result[ii+1] = elements[0..ii].to_s}
        result
      $ruby$ language plruby immutable;
      
      create or replace function xenodot.ancestor_hierarchy_set(val text) returns setof text as $pgsql$
        declare
          ii int4;
          ancestors text[];
          current text;
        begin
          ancestors := xenodot.ancestor_hierarchy(val);
          ii := 1;
          loop
            if array_upper(ancestors, 1) < ii then
              return;
            end if;
            return next ancestors[ii];
            ii := ii + 1;
          end loop;
          return;
        end;
      $pgsql$ language plpgsql immutable;
      
      create or replace function xenodot.child_position(text) returns int4 as $ruby$
        pos_encoded = args[0][/~*\w\Z/]
        return nil unless pos_encoded
        pos = 0
        pos_encoded.each_byte do |c|
          case c
          when 49..57: pos += c - 48
          when 65..90: pos += c - 55
          when 97..122: pos += c - 61
          when 126: pos += 61
          else
            return 0
          end
        end
        pos
      $ruby$ language plruby immutable;
      
      create or replace function xenodot.child_hierarchy(text, int4) returns text as $ruby$
        init_xenodot
        Xenodot::Backend::Hierarchy(*args)
      $ruby$ language plruby immutable;
    }

    execute <<-SQL
      create or replace function xenodot.root_node(uri text) returns xenodot.node as $pgsql$
        declare
          true_root_id int4;
          lock_helper int4;
          result xenodot.node;
        begin
          true_root_id := xenodot.ensure_root_id(uri);
          
          select * into result from xenodot.node where root_id = true_root_id and hierarchy = '';
          if not found then
            select id into lock_helper from xenodot.uri where id = true_root_id for share;
            select * into result from xenodot.node where root_id = true_root_id and hierarchy = '';
            if not found then
              insert into xenodot.node(root_id, hierarchy, name_id) values(true_root_id, '', -1) returning * into result;
            end if;
          end if;
          return result;
        end;
      $pgsql$ language plpgsql;
    SQL
    
    execute %q{
      create or replace function xenodot.last_modified_id() returns int4 as $sql$
        select -4;
      $sql$ language sql immutable;
      
      insert into xenodot.qname(id, ns, local_name) values(xenodot.last_modified_id(), '!xenodot!', 'last-modified');
      
      create or replace function xenodot.last_modified() returns xenodot.qname as $sql$
        select * from xenodot.qname where id = xenodot.last_modified_id();
      $sql$ language sql immutable;
    }
    
    execute %q{
      create or replace function xenodot.clean(root xenodot.node) returns void as $pgsql$
        begin
          delete from xenodot.node_head where root_id = root.root_id and hierarchy like root.hierarchy || '_%';
          delete from xenodot.property_head where node_id = root.id and name_id <> xenodot.last_modified_id();
          update xenodot.node set id = id where id = root.id; -- touch this node
        end;
      $pgsql$ language plpgsql;
    }
    
    execute <<-SQL
      create or replace function xenodot.clean_root(uri text) returns xenodot.node as $pgsql$
        declare
          result xenodot.node;
        begin
          select * into result from xenodot.root_node(uri);
          perform xenodot.clean(result);
          return result;
        end;
      $pgsql$ language plpgsql;
    SQL
    
    execute <<-SQL
      create type xenodot.value as (
        type "char",
        content bytea
      )
    SQL

    create_table 'xenodot.property_head' do |t|
      t.column  :node_id,       :int4,  :null => false
      t.column  :name_id,       :int4,  :null => false
      t.column  :position,      :int4,  :null => false, :default => 0
      t.column  :value,         'xenodot.value', :null => false
    end
    
    execute %q{
      create or replace view xenodot.property as select * from xenodot.property_head;
    }
    
    execute <<-SQL
      alter table xenodot.property_head add foreign key(node_id) references xenodot.node_head on delete cascade initially deferred;
      alter table xenodot.property_head add foreign key(name_id) references xenodot.qname initially deferred;

      alter table xenodot.property_head add constraint sane_position check(position >= 0);
      alter table xenodot.property_head add constraint sane_value 
        check((position = 0) or ((value).type is not null and (value).content is not null));
      
      create unique index property_access on xenodot.property_head(node_id, name_id, position);
      cluster property_access on xenodot.property_head;
      
      create index property_name on xenodot.property_head(name_id);
    SQL

    execute %q{
      create or replace function xenodot.create_child(xenodot.node, xenodot.qname) returns xenodot.node as $pgsql$
        declare
          parent_id int4;
          pos int4;
          result xenodot.node;
        begin
          select id into parent_id from xenodot.node_head where id = $1.id for share;
          select count(*)+1 into pos from xenodot.node_head where xenodot.parent_hierarchy(hierarchy) = $1.hierarchy;
          
          select * into result from xenodot.create_child($1, pos, $2);
          return result;
        end;
      $pgsql$ language plpgsql;
    }

    execute %q{
      create or replace function xenodot.create_child(xenodot.node, int4, xenodot.qname) returns xenodot.node as $pgsql$
        declare
          result xenodot.node;
        begin
          if $2 < 1 then
            select * from xenodot.create_child($1, $3) into result;
          else
            insert into xenodot.node(root_id, hierarchy, name_id) select $1.root_id, xenodot.child_hierarchy($1.hierarchy, $2), $3.id returning * into result;
          end if;
          return result;
        end;
      $pgsql$ language plpgsql;
    }
    
    execute %q{
      create or replace function xenodot.ensure_child(parent xenodot.node, name xenodot.qname) returns xenodot.node as $pgsql$
        declare
          result xenodot.node;
        begin
          select    * 
          from      xenodot.node 
          into      result 
          where     xenodot.parent_hierarchy(hierarchy) = parent.hierarchy
                and name_id = name.id
                and root_id = parent.root_id
          order by  hierarchy desc
          limit 1;
          
          if not found then
            result := xenodot.create_child(parent, name);
          end if;
          
          return result;
        end
      $pgsql$ language plpgsql;
    }
    
    execute %q{
      create or replace function xenodot.move_to(mover xenodot.node, new_parent xenodot.node) returns void as $pgsql$
        declare
          lock_keeper   xenodot.node;
          pos           int4;
        begin
          select * into lock_keeper from xenodot.node where id = new_parent.id for share;
          select count(*)+1 into pos from xenodot.node where xenodot.parent_hierarchy(hierarchy) = new_parent.hierarchy;
          
          update  xenodot.node
            set   root_id = new_parent.root_id,
                  hierarchy = xenodot.child_hierarchy(new_parent.hierarchy, pos)
          where   id = mover.id;
        end;
      $pgsql$ language plpgsql;
    }
    
    execute %q{
      create or replace function xenodot.move_left(mover xenodot.node, new_right xenodot.node) returns void as $pgsql$
        begin
          update xenodot.node
            set  hierarchy = xenodot.child_hierarchy(xenodot.parent_hierarchy(new_right.hierarchy), xenodot.child_position(new_right.hierarchy) + 1)
          where  id = new_right.id;
          
          update xenodot.node
            set  root_id = new_right.root_id,
                 hierarchy = new_right.hierarchy
          where  id = mover.id;
        end;
      $pgsql$ language plpgsql;
    }
    
    execute %q{
      create or replace function xenodot.move_right(mover xenodot.node, new_left xenodot.node) returns void as $pgsql$
        begin
          update xenodot.node
            set  root_id = new_left.root_id,
                 hierarchy = xenodot.child_hierarchy(xenodot.parent_hierarchy(new_left.hierarchy), xenodot.child_position(new_left.hierarchy) + 1)
          where  id = mover.id;
        end;
      $pgsql$ language plpgsql;
    }
    
    execute %q{
      create or replace function xenodot.clone_to(source xenodot.node, new_parent xenodot.node) returns xenodot.node as $pgsql$
        declare
          name      xenodot.qname;
          new_root  xenodot.node;
          row       record;
        begin
          if source.root_id = new_parent.root_id and new_parent.hierarchy like source.hierarchy || '%' then
            raise exception 'Recursion while cloning detected!';
          end if;
          
          -- create the new head (triggers last modified)
          select * into name from xenodot.qname where id = source.name_id;
          new_root := xenodot.create_child(new_parent, name);
          
          -- copy head's properties
          insert into xenodot.property_head(node_id, name_id, position, value)
            select new_root.id, name_id, position, value from xenodot.property_head where node_id = source.id;

          -- loop descendants
          for row in
            select nextval('xenodot.node_head_id_seq'::regclass) as target_id, node_head.*
            from      xenodot.node_head
            where     root_id = source.root_id
                  and hierarchy like source.hierarchy || '_%'
          loop
            insert into xenodot.node_head(id, root_id, hierarchy, name_id)
              values(row.target_id, new_root.root_id, overlay(row.hierarchy placing new_root.hierarchy from 1 for char_length(source.hierarchy)), row.name_id);
              
            insert into xenodot.property_head(node_id, name_id, position, value)
              select row.target_id, name_id, position, value
              from xenodot.property_head
              where node_id = row.id;
          end loop;
          
          return new_root;
        end;
      $pgsql$ language plpgsql;
    }

  end
  
  def self.down
  end
end