require 'yaml'

module Xenodot
  module Backend
    MUTEX = Mutex.new
    
    class Database
      @@connections = []
      
      def self.connect!(target = self)
        target.establish_connection(YAML.load_file(File.join(File.dirname(__FILE__), '..', '..', '..', 'config', 'database.yml'))[$XENODOT_ENV])
      end
      
      def self.establish_connection(yaml)
        @@config = {}
        @@config['host'] = yaml['host']
        @@config['port'] = yaml['port'] || 5432
        @@config['dbname'] = yaml['database']
        @@config['user'] = yaml['username']
        @@config['password'] = yaml['password']
        @@encoding = yaml['encoding'] || 'utf8'
        @@errorlevel = yaml['error_level'] || 'error'
      end
      
      def self.connection(options = {})
        MUTEX.synchronize do
          conn = Thread.current[:xenodot_connection] || @@connections.pop || begin
            PGconn.translate_results = true
            conn = PGconn.open(@@config)
            conn.async_exec("set client_encoding to #{@@encoding};")
            conn.async_exec("set client_min_messages to #{@@errorlevel};")
            return conn
          end
          Thread.current[:xenodot_connection] = conn if options[:transaction]
          return conn
        end
      end
      
      def self.return_connection(conn)
        unless Thread.current[:xenodot_connection] == conn
          MUTEX.synchronize do
            @@connections.push(conn)
          end
        end
      end
      
      def self.remove_connection
        MUTEX.synchronize do
          @@connections.each{|conn| conn.close}
          @@connections = []
        end
      end
      
      connect!
    end
    
    def self.sql(statement, values, types, options={})
      debug = options.key?(:debug) ? options[:debug] : $XENODOT_SHOW_SQL
      if debug
        warn ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>"
        warn "Query:  #{statement}"
        warn "Params: #{values.inspect}"
        warn "Types:  #{types.inspect}"
      end
      
      sql = statement.gsub(/\$(\d+)/) do
        offset = $1.to_i - 1
        case types[offset].to_sym
        when :bytea, :binary
          PGconn.escape_bytea(values[offset])
        else
          PGconn.quote(values[offset])
        end
      end
      warn "SQL:    #{sql}" if debug
      
      conn = Database.connection
      begin
        PGconn.translate_results = true
        result_raw = conn.async_exec(sql)
        result = nil
        output = (options[:output] || options['output'] || :value).to_sym

        if result_raw.result.size == 0
          return false
        elsif output == :hashed
          result = result_raw.result.collect do |row|
            hashed_row = Hash.new
            row.each_with_index do |value, index|
              hashed_row[result_raw.fieldname(index)] = value
            end
            hashed_row
          end
        else
          result = result_raw.result
        end
        result = result[0] if options[:count] == 1

        result_raw.clear()
        return result
      ensure
        if debug
          warn "Result: #{result.inspect}"
          warn "<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<"
        end
        Database.return_connection(conn)
      end
    end
    
    def self.transaction! &block
      if self.transaction?
        subtransaction! &block
      else
        begin
          Database.connection(:transaction => true)
          sql('begin;', [], [])
          break_called = true
          yield
          sql('commit;', [], [])
          break_called = false
        rescue => ex
          raise ex
        ensure
          sql('rollback;', [], []) if ex or break_called
          conn, Thread.current[:xenodot_connection] = Thread.current[:xenodot_connection], nil
          Database.return_connection(conn) if conn
        end
      end
    end
    
    def self.transaction?
      Thread.current[:xenodot_connection] != nil
    end

  end
end