module Xenodot
  module Backend
    module XmlParser
      
      class TempNode
        attr_reader :hierarchy
        attr_accessor :id
        attr_accessor :text
        
        def initialize(parent, position, name)
          if parent.kind_of? TempNode
            @hierarchy = Xenodot::Backend::Hierarchy(parent.hierarchy, position)
          else
            @hierarchy = Xenodot::Backend::Hierarchy(parent, position)
          end
          @name = name
          @id = nil
        end
        
        def sql(root_id)
          "(#{root_id}, '#{@hierarchy}', (xenodot.ensure_qname(E'#{PL.quote(@name.namespace_uri)}', E'#{PL.quote(@name.local_name)}')).id)"
        end
        
        def add_attr(name, position, value)
          (@attributes ||= []) << [name, position, value]
        end
        
        def attr_sql
          return nil unless @attributes
          @attributes.collect do |attr|
            attr_name, attr_pos, attr_value = attr
            value_func = case attr_value
            when URI:
              value_func = "xenodot.value_uri(xenodot.ensure_uri(E'#{PL.quote(attr_value.to_s)}'))"
            when HTree::Name:
              value_func = "xenodot.value_name(E'#{PL.quite(attr_value.namespace_uri)}', E'#{PL.quite(attr_value.local_name)}')"
            else
              value_func = "xenodot.value_string(E'#{PL.quote(attr_value)}')"
            end 
            "(#{@id}, (xenodot.ensure_qname(E'#{PL.quote(attr_name.namespace_uri)}', E'#{PL.quote(attr_name.local_name)}')).id, #{attr_pos}, #{value_func})"
          end.join(',')
        end
      end
      
      class Container
        def initialize
          @nodes = []
        end
        
        def add(parent, position, name)
          node = TempNode.new(parent, position, name)
          @nodes << node
          node
        end
        
        def add_text(parent, position, text)
          node = TempNode.new(parent, position, nil)
          node.text = text
          @nodes << node
          def node.sql(root_id)
            "(#{root_id}, '#{@hierarchy}', (xenodot.xml_meta_name()).id)"
          end
          def node.attr_sql
            "(#{@id}, (xenodot.xml_text_attr_name()).id, 0, xenodot.value_string(E'#{PL.quote(@text)}'))"
          end
        end
        
        def store(root_id)
          start_time = Time.now
          PL.exec("insert into xenodot.node(root_id, hierarchy, name_id) values #{@nodes.collect{|n| n.sql(root_id)}.join(',')} returning id;", 'output' => 'value').each_with_index do |row , offset|
            @nodes[offset].id = row[0]
          end
          node_time = Time.now
          warn "Nodes in #{node_time - start_time}"
          attributes = @nodes.collect{|n| n.attr_sql}.compact
          if attributes.size > 0
            PL.exec("insert into xenodot.property(node_id, name_id, position, value) values #{attributes.join(',')};")
          end
          end_time = Time.now
          warn "Attributes in #{end_time - node_time}, #{end_time - start_time} in total"
        end
      end
      
      CDATA = /\A<\!\[CDATA\[(.*)\]\]\>\Z/
      WHITESPACE_NAMES = %w{{http://www.w3.org/1999/xhtml}pre {http://www.w3.org/1999/xhtml}code code}

      def self.parse(value, root_id, head_hierarchy)
        if value.kind_of? URI
          base_uri = value
          warn "Fetching"
          value.open do |stream|
            base_uri = stream.base_uri
            value    = stream.read
            if stream.charset
              warn "Converting from #{stream.charset} to UTF-8"
              value  = Iconv.conv('UTF-8', stream.charset, value)
            end
            stream.meta.each do |key, val|
              #new_attr(root_id, :meta, key,
              #  case key
              #  when 'expires', 'date' then                  
              #    Time.rfc822(val)
              #  else
              #    val
              #  end
              #)
            end
          end
        else
          # TODO
          base_uri = URI.parse('')
        end
        warn "Parsing"
        dom = HTree(value)
        warn "Traversing"
        tree = Container.new        
        recurse(tree, dom, head_hierarchy, base_uri)
        warn "Writing"
        tree.store(root_id)
        warn "Done"
      end

      def self.recurse(tree, htree_node, parent, base_uri)
        node_pos = 0
        htree_node.each_child do |n|
          case n
          when HTree::Text:
            raw = n.raw_string.scan(CDATA)
            tree.add_text(parent, node_pos +=1, raw.length > 0 ? raw.to_s : n.to_s)
          when HTree::Elem:
            node = tree.add(parent, node_pos += 1, n.element_name)
            n.attributes.each do |qname, value|
              begin
                if qname.namespace_prefix == 'xml' and qname.local_name == 'base'
                  base_uri = base_uri + value.to_s
                end

                case qname.local_name
                when 'src', 'href':
                  begin
                    node.add_attr(qname, 0, base_uri + value.to_s)
                  rescue
                    node.add_attr(qname, 0, value.to_s)
                  end
                #when 'rel', 'rev':
                #    new_attr(node_id, qname, HTree::Name.parse_attribute_name(value.to_s, n.context))
                #when 'class':
                  # class stores qnames separated by whitespace
                  #classes = value.to_s.split.collect do |token|
                  #  QName.parse(*(token.scan(/()(\A-?[^_-]+\Z)|\A(-?[^_-]+)[_-](.*)\Z/)[0].compact + [n.context]))
                  #end

                  # erdf treats a prefixing "-" differently, so store the rest first
                  #dom[qname] = classes.reject{|c| c[0].slice(0,1) == '-'}.collect{|name| QName[*name]}

                  # everything else is really an rdf:type, let's store it that way
                  #rdf_types = classes.select{|c| c[0].slice(0,1) == '-'}
                  #dom[:rdf, :type] = rdf_types.collect{|type| QName[type[0][1..-1], type[1]]} if rdf_types.size > 0
#                when 'profile':
#                  value.to_s.split.each_with_index do |token, ii|
#                    new_attr(node_id, qname, begin base_uri + token rescue token end, ii+1)
#                  end
                else
                  node.add_attr(qname, 0, value.to_s)
                end
              rescue
                warn $!
                node.add_attr(qname, 0, value.to_s)
              end
            end
            recurse(tree, n, node, base_uri)
          when HTree::Comment
          when HTree::DocType
          when HTree::ProcIns
          end
        end
      end
    end
  end
end