module Xenodot
  module Query
    module XPath
      OPERATORS = {
        :eq => '=',
        :neq  => '<>',
        :lt => '<',
        :lteq => '<=',
        :gt => '>',
        :gteq => '>=',
      }
      
      OP_KEYS = OPERATORS.keys
      
      def self.build_query(start_node, path)
        parser = REXML::Parsers::XPathParser.new
        parser.namespaces = {}
        
        xpath = parser.parse(path)
        
        query = AbstractQuery.new(nil) do |sql|
          sql.from            = 'xenodot.node as current_node'
          if start_node
            sql.where         = 'current_node.id = ?'
            sql.where_params  = start_node['id']
            sql.where_types   = :int4
          end
        end
        
        parse_xpath(query, xpath)
      end

      def self.parse_xpath(query, xpath, sqlprefix = '')
        ii = 0;
        xpath = [xpath] unless xpath.kind_of? Array

        while (ii < xpath.size)
          case xpath[ii]
          when :qname:
            focus = query.sql?.default.to_s.gsub(/\.\*\Z/, '')
            query.sql?.from         = "#{query.sql?.from} join xenodot.qname(?, ?) as #{focus}_name on (#{focus}.name_id = #{focus}_name.id)"
            query.sql?.from_params  << xpath[ii += 1] << xpath[ii += 1]
            query.sql?.from_types   << :text << :text
          when :attribute:
            query.sql?.from         += ', xenodot.property as current_property'
            query.sql?.where        = [query.sql?.where, 'current_node.id = current_property.node_id'].join(' and ')
            query.sql?.default      = :'current_property.*'
            query.sql?.default_type = :'xenodot.property'
          when :predicate:
            raise "Predicates are allowed on nodes only!" unless query.sql?.default_type == :'xenodot.node'
            query = parse_predicate(query, xpath[ii += 1], "sub#{sqlprefix}")
          when :document:
            query = AbstractQuery.new(query) do |sql|
              sql.from              = 'xenodot.node as current_node'
              sql.where             = %q{current_node.root_id = parent1_node.root_id and current_node.hierarchy = ''}
            end
          when :node, :self:
            # do nothing
          when :child:
            query = AbstractQuery.new(query) do |sql|
              sql.from              = 'xenodot.node as current_node'
              sql.where             = 'current_node.root_id = parent1_node.root_id and xenodot.parent_hierarchy(current_node.hierarchy) = parent1_node.hierarchy'
            end
          when :parent:
            if query.sql?.default_type  ==  :'xenodot.property'
              query.sql?.default        =   :'current_node.*'
              query.sql?.default_type   =   :'xenodot.node'
            else
              query = AbstractQuery.new(query) do |sql|
                sql.from            = 'xenodot.node as current_node'
                sql.where           = 'current_node.root_id = parent1_node.root_id and xenodot.parent_hierarchy(parent1_node.hierarchy) = current_node.hierarchy'
              end
            end
          when :descendant_or_self:
            query = AbstractQuery.new(query) do |sql|
              sql.from            = 'xenodot.node as current_node'
              sql.where           = %q{current_node.root_id = parent1_node.root_id and current_node.hierarchy like parent1_node.hierarchy || '%'}
            end
            ii += 2 if xpath[ii+1] == :node and xpath[ii+2] == :child
          when :descendant:
            query = AbstractQuery.new(query) do |sql|
              sql.from              = 'xenodot.node as current_node'
              sql.where             = %q{current_node.root_id = parent1_node.root_id and current_node.hierarchy like parent1_node.hierarchy || '_%'}
            end            
          when :ancestor_or_self:
            query = AbstractQuery.new(query) do |sql|
              sql.from              = 'xenodot.node as current_node'
              sql.where             = 'current_node.root_id = parent1_node.root_id and current_node.hierarchy = ANY(xenodot.ancestor_hierarchy(parent1_node.hierarchy))'
            end
          when :ancestor:
            query = AbstractQuery.new(query) do |sql|
              sql.from              = 'xenodot.node as current_node'
              sql.where             = 'current_node.root_id = parent1_node.root_id and current_node.hierarchy = ANY(xenodot.ancestor_hierarchy(parent1_node.hierarchy)) and current_node.id <> parent1_node.id'
            end
          when :following:
      			query = AbstractQuery.new(query) do |sql|
              sql.from              = 'xenodot.node as current_node'
              sql.where             = %q{current_node.root_id = parent1_node.root_id and current_node.hierarchy not like (parent1_node.hierarchy || '%') and current_node.hierarchy > parent1_node.hierarchy}
            end
          when :following_sibling:
            query = AbstractQuery.new(query) do |sql|
              sql.from              = 'xenodot.node as current_node'
              sql.where             = 'current_node.root_id = parent1_node.root_id and xenodot.parent_hierarchy(current_node.hierarchy) = xenodot.parent_hierarchy(parent1_node.hierarchy) and current_node.hierarchy > parent1_node.hierarchy'
            end
          when :preceding:
            query = AbstractQuery.new(query) do |sql|
              sql.from              = 'xenodot.node as current_node'
              sql.where             = 'current_node.root_id = parent1_node.root_id and current_node.hierarchy <> ALL(xenodot.ancestor_hierarchy(parent1_node.hierarchy)) and current_node.hierarchy < parent1_node.hierarchy'
            end
          when :preceding_sibling:
            query = AbstractQuery.new(query) do |sql|
              sql.from              = 'xenodot.node as current_node'
              sql.where             = 'current_node.root_id = parent1_node.root_id and xenodot.parent_hierarchy(current_node.hierarchy) = xenodot.parent_hierarchy(parent1_node.hierarchy) and current_node.hierarchy < parent1_node.hierarchy'
            end            
          else
            raise "Don't know path #{xpath.inspect}"
          end          
          ii += 1
        end
        query
      end
      
      def self.parse_predicate(query, xpath, sqlprefix)
        xpath = compact_predicate_tree(xpath)
        xpath, ignore = optimize_predicate_tree(xpath)
        where, params, types = build_predicate(query, xpath, sqlprefix)
        query.sql?.where << " and (#{where})"
        query.sql?.where_params += params
        query.sql?.where_types += types
        return query
      end
      
      # this is just a simple compactification, recurse and/or are merged
      def self.compact_predicate_tree(xpath)
        case xpath[0]
        when :and, :or
          operator = xpath[0]
          new_xpath = [operator]
          xpath[1..-1].each do |element|
            element = compact_predicate_tree(element)
            if element[0] == operator then
              new_xpath += element[1..-1]
            else
              new_xpath << element
            end
          end
          new_xpath
        when :function
          if xpath[1] == 'position'
            [:position]
          elsif xpath[1] == 'rposition' # reversed position
            [:rposition]
          elsif xpath[1] == 'not'
            new_xpath = [:not]
            xpath[2].each do |negate|
              new_xpath << compact_predicate_tree(negate)
            end
            new_xpath
          else
            new_xpath = xpath[0..1]
            xpath[2].each do |parameter|
              new_xpath << compact_predicate_tree(parameter)
            end
            new_xpath
          end
        when *OP_KEYS:
          new_xpath = [xpath[0]]
          xpath[1..-1].each do |parameter|
            new_xpath << compact_predicate_tree(parameter)
          end
          new_xpath
        when :neg
          raise "Panic on #{xpath.inspect}" unless xpath[1] == :literal
          [:literal, -xpath[2]]
        else
          xpath
        end
      end
      
      # this splits up the tree into correct statements
      # xpath-or can not easily be matched to sql-or, thus we need to split it
      # Opimization TODO: http://en.wikipedia.org/wiki/Disjunctive_normal_form
      def self.optimize_predicate_tree(xpath, existance = true)
        case xpath[0]
        when :and
          new_xpath = [:and]
          complex_part = Set.new
          xpath[1..-1].each do |element|
            sub_xpath, sub_complex_part = optimize_predicate_tree(element)
            new_xpath << sub_xpath
            complex_part += sub_complex_part
          end
          if complex_part.find{|n| n[0].to_s =~ /^outter/}
            new_xpath[0] = :outter_and
            complex_part = Set.new([[:outter]])
          end
          return new_xpath, complex_part
        when :or
          new_xpath = [:or]
          complex_part = Set.new
          xpath[1..-1].each do |element|
            sub_xpath, sub_complex_part = optimize_predicate_tree(element)
            new_xpath << sub_xpath
            complex_part += sub_complex_part
          end
          if complex_part.find{|n| n[0].to_s =~ /^outter/} or complex_part.size > 1
            new_xpath[0] = :outter_or 
            complex_part = Set.new([[:outter]])
          end
          return new_xpath, complex_part
        when :literal
          return xpath, Set.new
        when :function
          complex_part = Set.new
          xpath[1..-1].each do |element|
            sub_xpath, sub_complex_part = optimize_predicate_tree(element, false)
            complex_part += sub_complex_part
          end
          return xpath, complex_part
        when :not
          new_xpath = [:not]
          sub_xpath, complex_part = optimize_predicate_tree(xpath[1])
          new_xpath << sub_xpath
          if complex_part.find{|n| n[0].to_s =~ /^outter/}
            new_xpath[0] = :outter_not
            complex_path = Set.new([[:outter]])
          end
          return new_xpath, complex_part
        when *OP_KEYS
          complex_part = Set.new
          xpath[1..-1].each do |element|
            sub_xpath, sub_complex_part = optimize_predicate_tree(element, false)
            complex_part += sub_complex_part
          end
          return xpath, complex_part
        else
          if existance
            return [:exists, xpath], Set.new([[:outter]])
          else
            return xpath, Set.new([xpath])
          end
        end
      end
      
      def self.build_predicate(query, xpath, sqlprefix)
        case xpath[0]
        when :outter_and, :outter_or
          sides = xpath[1..-1].collect do |side|
            build_predicate(query, side, sqlprefix)
          end.transpose
          
          return '(' + sides[0].collect{|side_query| "(#{side_query})"}.join(" #{xpath[0].to_s.gsub(/^.+_/, '')} ") + ')',
            sides[1].flatten, sides[2].flatten
        when :outter_not
          from, params, types = build_predicate(query,xpath[1], sqlprefix)
          return "not(#{from})", params.flatten, types.flatten
        when :literal
          return Xenodot.encode(xpath[1]), [], [] # todo: turn into boolean
        when :and, :or, :function, :position, :rposition, :not, *OP_KEYS
          complex = build_complex_parts(query, xpath, sqlprefix)
          selection = build_predicate_selection(xpath, complex)
          complex.each{|key, value| value[1][0] << " as #{value[0]}"}
          complex = complex.collect{|key, value| value[1]}.transpose
          froms = complex[0].flatten
          params = complex[1].flatten
          types = complex[2].flatten
          
          return "select #{selection} from #{froms.join(',')}", params, types
        when :exists
          sub_query, params, types = build_complex_parts(query, xpath[1], sqlprefix)[xpath[1]][1]
          return "exists#{sub_query}", params.flatten, types.flatten
        else
          raise "Panic on parse: #{xpath.inspect}"
        end
      end
      
      def self.build_predicate_selection(xpath, complex, wrap = false)
        case xpath[0]
        when :literal
          return Xenodot.encode(xpath[1])
        when *OP_KEYS
          return '(' + xpath[1..-1].collect do |op| 
            build_predicate_selection(op, complex, true)
          end.join(" #{OPERATORS[xpath[0]]}") + ')'
        when :and, :or
          return '(' + xpath[1..-1].collect do |side|
            build_predicate_selection(side, complex)
          end.join(" #{xpath[0]} ") + ')'
        when :not
          return "not(#{build_predicate_selection(xpath[1], complex)})"
        when :function
          return "xpath.#{xpath[1].gsub(/-/, '_')}(" + xpath[1..-1].collect do |param|
            build_predicate_selection(param, complex, false)
          end.join(',') + ')'
        when :position, :rposition
          return "#{complex[xpath][0]}.pos"
        else
          if wrap
            "xpath.string(cast(#{complex[xpath][0]} as #{complex[xpath][1][3]}))"
          else
            complex[xpath][0]
          end
        end
      end
      
      def self.build_complex_parts(query, xpath, sqlprefix)
        sub_queries = {}
        scan_leaf_predicate(xpath).each_with_index do |sub_xpath, pos|
          sub_prefix = "#{sqlprefix}#{pos}"
          sub_queries[sub_xpath] = sub_prefix, case sub_xpath[0]
          when :position
            sub_query = AbstractQuery.new(nil) do |sql|
              sql.from = "xenodot.node as current_node"
              sql.where = %Q{    #{query.sql?.where.gsub(/parent1\_node/, 'TRUEPARENT')} 
                 and (    (current_node.root_id < TRUEROOT.root_id) 
                      or  (current_node.root_id = TRUEROOT.root_id and current_node.hierarchy < TRUEROOT.hierarchy)
                    )}
            end.sql!(:no_exec => true, :prefix => sub_prefix)
            # assignment to sub_queries[sub_xpath][1]
            [
              "(select count(*)+1 as pos from #{sub_query[:froms].join(',')} where #{sub_query[:wheres].join(' and ')})",
              sub_query[:from_params] + sub_query[:where_params],
              sub_query[:from_types] + sub_query[:where_types],
              sub_query[:default_type]
            ]
          when :rposition
            sub_query = AbstractQuery.new(nil) do |sql|
              sql.from = "xenodot.node as current_node"
              sql.where = %Q{    #{query.sql?.where.gsub(/parent1\_node/, 'TRUEPARENT')} 
                 and (    (current_node.root_id > TRUEROOT.root_id) 
                      or  (current_node.root_id = TRUEROOT.root_id and current_node.hierarchy > TRUEROOT.hierarchy)
                    )}
            end.sql!(:no_exec => true, :prefix => sub_prefix)
            # assignment to sub_queries[sub_xpath][1]
            [
              "(select -(count(*)+1) as pos from #{sub_query[:froms].join(',')} where #{sub_query[:wheres].join(' and ')})",
              sub_query[:from_params] + sub_query[:where_params],
              sub_query[:from_types] + sub_query[:where_types],
              sub_query[:default_type]
            ]
          else
            sub_query = parse_xpath(AbstractQuery.new(nil) do |sql|
              sql.from = "xenodot.node as current_node"
              sql.where = "current_node.id = TRUEROOT.id"
            end, sub_xpath, sub_prefix).sql!(:no_exec => true, :prefix => sub_prefix)
            # assignment to sub_queries[sub_xpath][1]
            [
              "(select #{sub_query[:default]} from #{sub_query[:froms].join(',')} where #{sub_query[:wheres].join(' and ')})",
              sub_query[:from_params] + sub_query[:where_params],
              sub_query[:from_types] + sub_query[:where_types],
              sub_query[:default_type]
            ]
          end
        end
        sub_queries
      end
      
      def self.scan_leaf_predicate(xpath, result = Set.new)
        case xpath[0]
        when :and, :or, *OP_KEYS
          xpath[1..-1].each{|n| scan_leaf_predicate(n, result)}
        when :function
          xpath[2..-1].each{|n| scan_leaf_predicate(n, result)}
        when :literal
          # do nothing
        when :not
          scan_leaf_predicate(xpath[1], result)
        else
          result << xpath
        end
        return result
      end
      
    end
  end
end