describe Xenodot::Node do
  before(:each) do
    reset_xenodot_database
    @uri  = 'http://localhost:3000/'
    @node = Xenodot::Node[@uri]
  end
  
  it "should return a persistent root node for URIs" do
    @node.id.should == Xenodot::Node[@uri].id
  end
    
  it "should ignore protocol for authorative URIs" do
    host = 'localhost:3000' # authorative by rake migrate
    Xenodot::Node['http://' + host].id.should == Xenodot::Node['https://' + host].id
  end
  
  it "should not ignore protocol for non-authorative URIs" do
    non_auth_host = 'localhost' # not on default port, hence not authorative
    Xenodot::Node['http://' + non_auth_host].id.should_not == Xenodot::Node['https://' + non_auth_host].id
  end
  
  it "should return nil for empty attributes" do
    @node[:foo].should == nil
  end
  
  it "should store Integers as attributes" do
    @node[:foo] = 1
    @node[:foo].class.should == Fixnum
    @node[:foo].should == 1
  end
  
  it "should store Strings as attributes" do
    @node[:foo] = 'hello'
    @node[:foo].class.should == String
    @node[:foo].should == 'hello'
  end
  
  it "should store QNames as attributes" do
    @node[:foo] = Xenodot::QName[:bar, :baz]
    @node[:foo].class.should == Xenodot::QName
    @node[:foo].should == Xenodot::QName[:bar, :baz]
  end
  
  it "should store references to other nodes" do
    reference = Xenodot::Node['http://somewhere.else.org']
    @node[:foo] = reference
    @node[:foo].class.should == Xenodot::Node
    @node[:foo].id.should == reference.id
    @node[:foo].should == reference
  end

  it "should store XPathBuilder" do
    reference = Xenodot::Node['http://somewhere.else.org']
    xpath = reference.child.following_sibling(:bar).parent
    @node[:foo] = xpath
    @node[:foo].class.should == Xenodot::XPathBuilder
    @node[:foo].start_node_id.should == reference.id
    @node[:foo].xpath.should == "child::/following-sibling::bar/parent::"
  end
  
  it "should store URIs" do
    uri = URI.parse('http://somewhere.else.org/a/b?c=d#f')
    @node[:foo] = uri
    @node[:foo].class.should == URI::HTTP
    @node[:foo].should == uri
  end
  
  it "should store Time" do
    now = Time.parse(Time.now.rfc822) # not storing millisecs
    @node[:foo] = now
    @node[:foo].class.should == Time
    @node[:foo].should == now
  end
  
  it "should store Boolean" do
    @node[:foo] = false
    @node[:foo].should == false
    @node[:foo] = true
    @node[:foo].should == true
  end
  
  it "should store Float" do
    f = 10.01
    @node[:foo] = f
    @node[:foo].class.should == Float
    @node[:foo].should == f
  end

  it "should store Arrays" do
    values = [1, '2', 3.04]
    @node[:foo] = values
    @node[:foo].class.should == Array
    @node[:foo].size.should == values.size
    @node[:foo].should == values
  end
  
  it "should store empty Arrays" do
    @node[:foo] = []
    @node[:foo].class.should == Array
    @node[:foo].size.should == 0
  end
  
  it "should have detached Arrays" do
    values = [1,2,3]
    @node[:foo] = values
    stored = @node[:foo]
    stored << 4
    @node[:foo].should == values # stored value is still values, not the modified read array
  end
  
  it "should allow removable of attributes by setting nil" do
    @node[:foo] = 1
    @node.attributes.size.should == 1
    @node.attributes.include?(Xenodot::QName[:foo]).should == true
    @node[:foo] = nil
    @node.attributes.size.should == 0
  end
  
  it "should have namespaced attributes" do
    ns = URI.parse('http://some.namespace.com/a/b/c?d=e#f')
    @node[ns, :foo] = 1
    @node[:foo].should == nil
    @node[ns, :foo].should == 1
    @node[ns.prefix, :foo].should == 1
    @node[ns.prefix.to_sym, :foo].should == 1
    @node[ns, 'foo'].should == 1
  end
    
  it "should allow access by prefix and namespace, as well as QName" do
    ns = URI.parse('http://some.namespace.com/a/b/c?d=e#f')
    @node[ns, :foo] = 1
    @node[Xenodot::QName[ns, :foo]].should == 1
    @node[Xenodot::QName[ns, 'foo']].should == 1
    @node[Xenodot::QName[ns.prefix, 'foo']].should == 1
    @node[Xenodot::QName[ns.prefix.to_sym, 'foo']].should == 1
  end
  
  it "should be able to create named child nodes" do
    ns = URI.parse('http://some.namespace.com/a/b/c?d=e#f')
    child = @node.create_child(ns, :foo)
    child.parent.id.should == @node.id
    @node.child.size.should == 1
    child.name.should == Xenodot::QName[ns, :foo]
  end

  it "should be able to ensure named child nodes" do
    ns = URI.parse('http://some.namespace.com/a/b/c?d=e#f')
    child1 = @node.ensure_child(ns, :foo)
    child1.parent.id.should == @node.id
    @node.child.size.should == 1
    child1.name.should == Xenodot::QName[ns, :foo]
    
    child2 = @node.ensure_child(ns, :foo)
    child2.should == child1
    @node.child.size.should == 1
    
    child1 = @node.ensure_child(:bar)
    child1.parent.id.should == @node.id
    @node.child.size.should == 2
    child1.name.should == Xenodot::QName[:bar]
    
    child2 = @node.ensure_child(:bar)
    child2.should == child1
    @node.child.size.should == 2
    
    child3 = @node.create_child(:process)
    child3[:pi] = child1.materialize!
    child3.ensure_child(:names)
    child3.child.name.should == Xenodot::QName[:names]

  end
  
  it "should always return the last child when ensuring" do
    ns = URI.parse('http://some.namespace.com/a/b/c?d=e#f')
    @node.child(ns, :foo).size.should == 0
    child1 = @node.ensure_child(ns, :foo)
    child2 = @node.ensure_child(ns, :foo)
    
    @node.child(ns, :foo).size.should == 1
    child1.should == child2
    
    child3 = @node.create_child(ns, :foo)
    @node.child(ns, :foo).size.should == 2    
    child4 = @node.ensure_child(ns, :foo)
    
    child3.should_not == child1
    child3.should == child4
  end
  
  it "should return deleted nodes on delete!" do
    child1 = @node.create_child(:foo)
    @node.child(:bar).size.should == 0
    child2 = @node.create_child(:bar)
    child3 = @node.create_child(:bar)
    @node.child(:bar).size.should == 2
    @node.child(:foo).delete!.should == child1
    @node.child(:bar).size.should == 2
    @node.child(:bar).delete!.should == [child2, child3]
  end

  it "should support integers on queries" do
    child1 = @node.create_child(:foo)
    child2 = @node.create_child(:foo)
    child3 = @node.create_child(:foo)
    
    @node.child(:foo).size.should == 3
    @node.child(:foo)[0].id.should == child1.id
    @node.child(:foo)[1].id.should == child2.id
    @node.child(:foo)[2].id.should == child3.id
    @node.child(:foo)[-1].id.should == child3.id
    @node.child(:foo)[-2].id.should == child2.id
    @node.child(:foo)[-3].id.should == child1.id
  end
  
  it "should support ranges on queries" do
    child1 = @node.create_child(:foo)
    child2 = @node.create_child(:foo)
    child3 = @node.create_child(:foo)
    
    @node.child(:foo).size.should == 3
    @node.child(:foo)[0..1].id.should == [child1.id, child2.id]
    @node.child(:foo)[1..2].id.should == [child2.id, child3.id]
    @node.child(:foo)[0..2].id.should == [child1.id, child2.id, child3.id]
    @node.child(:foo)[0...2].id.should == [child1.id, child2.id]
    @node.child(:foo)[1..-1].id.should == [child2.id, child3.id]
    @node.child(:foo)[1..-2].id.should == child2.id
    @node.child(:foo)[-3...-1].id.should == [child1.id, child2.id]
    @node.child(:foo)[-3..-1].id.should == [child1.id, child2.id, child3.id]
  end
  
  it "should be able to get a new parent" do
    par1 = @node.create_child(:par1)
    par2 = @node.create_child(:par2)
    child = par1.create_child(:chil)
    
    child.parent.id.should == par1.id
    child.parent = par2
    child.parent.id.should == par2.id
    par1.child.exist?.should == false
    par2.child.id.should == child.id
  end
  
	it "shcould be able to copy nodes" do
		par1 = @node.create_child(:par1)
		par2 = @node.create_child(:par2)
		par3 = @node.create_child(:par3)
		child = par1.create_child(:chil)
		child2 = child.create_child(:chil2)
		child2[:foo] = @node
		
		# should work with nodes
		par2.append_clone(child)
		
		par1.child.id.should == child.id
		par1.child.child.id.should == child2.id
		par1.child.child[:foo].should == @node
		par2.child.name.should == child.name
		par2.child.id.should_not == child.id
		par2.child.child.name.should == child2.name
		par2.child.child.id.should_not == child.id
		par2.child.child[:foo].should == @node
	
		# should work with XPathBuilder
		@node.child(:par3).append_clone(par1.child)
		
		par1.child.id.should == child.id
		par1.child.child.id.should == child2.id
		par1.child.child[:foo].should == @node
		par3.child.name.should == child.name
		par3.child.id.should_not == child.id
		par3.child.child.name.should == child2.name
		par3.child.child.id.should_not == child.id
		par3.child.child[:foo].should == @node

	end

  
  it "should be able to append xml" do
  	@node.append_xml('<test par="1"><child1 xmlns="http://xenodot.org/test" name="bla"/><child2 xmlns="http://xenodot.org/test2" name="blo"/></test>')
  	test = @node.child
  	test.name.should == Xenodot::QName[:test]
  	test[:par].should == '1'
  	test.child.size.should == 2
  	test.child[0].name.should == Xenodot::QName['http://xenodot.org/test',:child1]
  	test.child[0][:name].should == 'bla'
  	test.child[1].name.should == Xenodot::QName['http://xenodot.org/test2',:child2]
  	test.child[1][:name].should == 'blo'
  end
  
  it "should be able to append xml w/o attributes" do
  	@node.append_xml("<test/>")
  	@node.child.name.local_name.should == "test"
  end
  
  it "should be able to be converted to xml" do
  	xml = '<test par="1"><ns9:child1 xmlns:ns9="http://xenodot.org/test" name="bla"></ns9:child1><ns10:child2 xmlns:ns10="http://xenodot.org/test2" name="blo"></ns10:child2></test>'
  	@node.append_xml(xml)
  	test = @node.child
  	test.to_xml.should == xml
  	test.to_s.should == xml
  end
	
end