#
# Copyright (C) 2008  Olaf Märker
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

include Xenodot

describe "A π-Process" do
	before(:each) do
		reset_pivm_database
		#reset_xenodot_database
		#Backend.sql('select pivm.vm_init()',[],[])
		Xenodot::Backend.sql('set client_min_messages to error;', [], [])
		PiVM.init
		PiVM.load(File.join(File.dirname(__FILE__),"pi_spec.xml"))
	end

	it "should be deleted on inaction" do
		pi = PiVM.agents.child[0].descendant("inaction")
		p = PiVM.create_process(pi)
		p[:pi].should == pi
		PiVM.handle_process(p)
		Backend.sql("select * from xenodot.node where id = $1",[p.id],[:int4]).should be_false
	end
	
	it "should retrieve unique names on fresh" do
		pi = PiVM.agents.child[0].descendant("fresh")
		p = PiVM.create_process(pi)
		p.child(:names)['a'].should be_nil
		PiVM.handle_process(p)
		p.child(:names)['a'].should =~ /a_.+/
		p[:pi].should == pi.child
	end

	
	it "should be able to communicate with another process" do
		pi1 = PiVM.agents.child[1].child
		p1 = PiVM.create_process(pi1)
		pi2 = PiVM.agents.child[2].child
		p2 = PiVM.create_process(pi2)
		PiVM.handle_process(p1) #read a(x)
		Backend.sql("select * from pivm.vm_queue where process_id = $1",[p1.id],[:int4]).should_not be_false
		PiVM.handle_process(p2) #fresh y
		p2.substitute('y').should =~ /y_.+/
		PiVM.handle_process(p2) #write a<y>
		p1 = PiVM::Queue.dequeue
		p1.substitute('x').should =~ /y_.+/
		PiVM.handle_process(p1) #fresh b
		p1.substitute('b').should =~ /b_.+/
		PiVM.handle_process(p1) #write x<b>
		PiVM.handle_process(p2) #read y(z)
		p2.substitute('z').should =~ /b_.+/
	end
	
	it "should be able to call another agent" do
		pi = PiVM.agent('T').start
		p = PiVM.create_process(pi)
		PiVM.handle_process(p)
		p[:pi].should == PiVM.agent('P').start
	end
	
	it "should be able to call another agent with parameters" do
		pi = PiVM.agent('ComplexCall1').start
		p = PiVM.create_process(pi)
		
		PiVM.handle_process(p) #fresh
		a = p.substitute('a')
		b = p.substitute('b')
		PiVM.handle_process(p) # call
		p[:pi].should == PiVM.agent('ComplexCall2').start
		p.substitute('x').should == a
		p.substitute('y').should == b
	end

	it "should allow replications" do
		pi1 = PiVM.agents.child.detect{|n| n[:name] == "S"}.child
		p1 = PiVM.create_process(pi1)
		p2 = PiVM.create_process(pi1)
		p2.id.should_not == p1.id
		p2[:pi].should == p1[:pi]
		pi_rep = PiVM.agents.child.detect{|n| n[:name] == "U"}.child
		rep = PiVM.create_process(pi_rep)
		
		pi_u_fresh = pi_rep.child.child
		
		# handle the restriction
		PiVM.handle_process(p1)
		PiVM.handle_process(p2)

		pi_s_read = p1[:pi].child
		# enqueue p1 and p2
		PiVM.handle_process(p1)
		PiVM.handle_process(p2)
		res = Backend.sql('select count(*) from pivm.vm_queue',[],[], :count => 1)
		res[0].should == 2
		
		# handle the replication
		PiVM.handle_process(rep)
		
		# check that all processes point to the correct process step
		PiVM.processes.child.each_with_index do |p,i|
			p[:pi].should == pi_s_read if i == 0 or i == 1
			p[:pi].should == pi_rep if i == 2
			p[:pi].should == pi_u_fresh if i == 3 or i == 4
		end
		
		p3 = PiVM.create_process(pi1)
		PiVM.handle_process(p3)
		PiVM.handle_process(p3)
		p3[:pi].should == pi_s_read
		
	end
	
	it "should allow parallel splits" do
		p_par = PiVM.agents.child.detect{|n| n[:name] == "PAR"}.child
		p_write = p_par.descendant("write")
		p_read = p_par.descendant("read")
		p_tau = p_par.descendant("tau")
		p = PiVM.create_process(p_par)
		p.substitute!("a","b")
		PiVM.handle_process(p)
		PiVM.processes.child.size.should == 3
		PiVM.processes.child.each do |child|
			child.act_as PiVM::Process
			child.substitute("a").should == "b"
			#p 'p_write = ' << p_write.inspect
			case child[:pi].id
			when p_write.id
				p_write = -1
			when p_read.id
				p_read = -1
			when p_tau.id
				p_tau = -1
			else raise "Failure: Child is not as expected"
			end
		end
		res = Backend.sql("select * from pivm.vm_queue",[],[])
		res.size.should == 2
	end
	
	it "should allow matches" do
		p_match = PiVM.agents.child.detect{|n| n[:name] == "MATCH"}.child
		p = PiVM.create_process(p_match)
		p.substitute!("a","b")
		PiVM.handle_process(p)
		p[:pi].name.local_name.should == "tau"
		PiVM.handle_process(p)
		p[:pi].name.local_name.should == "call"
		PiVM.handle_process(p)
		p.substitute!("a",nil)
		PiVM.handle_process(p).should be_false
		p.exist?.should be_false
	end
	
	it "should allow summations" do
		p_sum = PiVM.agents.child.detect{|n| n[:name] == "SUM"}.child
		p = PiVM.create_process(p_sum)
		PiVM.handle_process(p) # fresh a
		PiVM.handle_process(p) # sum: should choose the tau
		p[:pi].name.local_name.should == "tau"
		Backend.sql('select * from pivm.vm_queue where process_id = $1',[p.id],[:int4],:count => 1).should be_false
		PiVM.handle_process(p) # tau
		PiVM.handle_process(p).should be_false # sum: should enqueue the read and the write
		ch = p.substitute("a")
		Backend.sql('select operation,channel from pivm.vm_queue where process_id = $1',[p.id],[:int4]).should == [['write', ch],['read',ch]]
		
	end
end
